>{-# OPTIONS -fglasgow-exts #-}

> module StreamArrow(
>	StreamArrow, StreamMap,
>	StreamMapST, runStream
> ) where

> import LazyST

> import Arrow
> import ArrowTransformer
> import ArrowOps
> import ArrowCircuit
> import Stream

> newtype StreamArrow a b c = Str (a (Stream b) (Stream c))
> unStr (Str x) = x

> instance Arrow a => Arrow (StreamArrow a) where
>	arr f = Str (arr (fmap f))
>	Str f >>> Str g = Str (f >>> g)
>	first (Str f) = Str (arr unzipStream >>> first f >>> arr zipStream)

> instance Arrow a => GenFunctor a Stream where
>	genmap f = arr (\xs -> (shd xs, stl xs)) >>>
>			f *** genmap f >>> arr (uncurry Cons)

Caution: genmap is only a functor if *** for the base arrow is.
(For Kleisli arrows, that would mean a commutative monad.)
The same goes for the equivalent lift: it can be used to lift arrows,
but won't preserve composition unless *** does.

> instance Arrow a => ArrowTransformer (StreamArrow) a where
>	lift f = Str (genmap f)

> instance ArrowFunctor (StreamArrow) where
>	arrowmap embed (Str f) = Str (embed f)

The following promotions follow directly from the arrow transformer.

> instance (ArrowZero a) => ArrowZero (StreamArrow a) where
>	zeroArrow = lift zeroArrow

> instance (ArrowExcept ex a) => ArrowExcept ex (StreamArrow a) where
>       raise = lift raise

> instance (ArrowReader r a) => ArrowReader r (StreamArrow a) where
>       readState = lift readState

> instance (ArrowState s a) => ArrowState s (StreamArrow a) where
>       fetch = lift fetch
>       store = lift store

> instance (ArrowWriter w a) => ArrowWriter w (StreamArrow a) where
>       write = lift write

liftings of standard classes

> instance Arrow a => ArrowChoice (StreamArrow a) where
>	left (Str f) = Str ((arr getLeft >>> f) &&& arr id >>> arr replace)
>		where	getLeft (Cons (Left x) xs) = Cons x (getLeft xs)
>			getLeft (Cons (Right _) xs) = getLeft xs
>			replace (~(Cons x xs), Cons (Left _) ys) =
>				Cons (Left x) (replace (xs, ys))
>			replace (xs, Cons (Right y) ys) =
>				Cons (Right y) (replace (xs, ys))

> instance ArrowLoop a => ArrowLoop (StreamArrow a) where
>	loop (Str f) = Str (loop (arr zipStream >>> f >>> arr unzipStream))

> instance ArrowPlus a => ArrowPlus (StreamArrow a) where
>	Str f <+> Str g = Str (f <+> g)

I don't know of any other useful promotions.
(runWriter can be promoted, but doesn't seem useful.)

Circuits

> instance ArrowLoop a => ArrowCircuit (StreamArrow a) where
>	delay x = Str (arr (Cons x))

> instance ArrowLoop a => ArrowRunCircuit (StreamArrow a) a where
>	runCircuit (Str f) = f

> type StreamMap = StreamArrow (->)

In-place state updates.

> type StreamMapST s = StreamArrow (Kleisli (ST s))

NB: this is an arrow type, and lift can be used to promote arrows from
Kleisli (ST s): the resulting arrow updates the state for each stream
element in turn, and as long as the final state in not required all is
well.  However, lift does not preserve composition, because this monad
isn't commutative (see the definition of lift above).  In particular,
a composition of lifts of state transformers will not work, as the second
will require the final state of the first.

Do something outside the stream and distribute the result.

> instance Arrow a => ArrowOnce StreamArrow a where
>	once f = Str (arr (\_ -> ()) >>> f >>> arr mkStream)
>		where	mkStream b = let bs = Cons b bs in bs

Encapsulation

> runStream :: (forall s. StreamMapST s b c) -> StreamMap b c
> runStream cf =
>	let Str f = cf in
>	let Kleisli g = f in
>	Str $ \input -> runST (g input)
